from __future__ import annotations

from typing import Any, Optional

from drepr.models.sm import DREPR_URI
from drepr.writers.base import StreamClassWriter
from rdflib import RDF, BNode, Graph, Literal, Namespace, URIRef

SubjVal = str | tuple | int | bool


class RDFGraphWriter(StreamClassWriter):
    def __init__(self, prefixes: dict[str, str]):
        self.g: Graph = Graph()
        for prefix, ns in prefixes.items():
            self.g.bind(prefix, Namespace(ns))

        self.written_records: dict[SubjVal, BNode | URIRef] = {}
        self.origin_subj: SubjVal = ""
        self.subj: Optional[URIRef | BNode] = None
        self.buffer: list[tuple[URIRef, URIRef | BNode | Literal]] = []
        self.is_buffered: bool = False
        # whether the current subject has any data or not
        # if the subject has URI, then it has data
        self.subj_has_data: bool = False

    def has_written_record(self, subj: SubjVal) -> bool:
        return subj in self.written_records

    def begin_record(
        self, class_uri: str, subj: SubjVal, is_blank: bool, is_buffered: bool
    ):
        self.origin_subj = subj
        if is_blank:
            if subj in self.written_records:
                self.subj = self.written_records[subj]
                self.subj_has_data = True
            else:
                self.subj = BNode()
                self.subj_has_data = False
        else:
            # subj will be a string for URIRef
            self.subj = URIRef(subj)  # type: ignore
            self.subj_has_data = True

        if is_buffered:
            self.buffer = [(RDF.type, URIRef(class_uri))]
        else:
            self.g.add((self.subj, RDF.type, URIRef(class_uri)))
        self.is_buffered = is_buffered
        self.subj_has_data = False

    def end_record(self):
        if self.subj is None:
            # has been aborted
            return

        if len(self.buffer) > 0:
            for pred, obj in self.buffer:
                self.g.add((self.subj, pred, obj))
            self.buffer = []
        self.written_records[self.origin_subj] = self.subj
        self.subj = None
        self.subj_has_data = False

    def abort_record(self):
        """Abort the record that is being written"""
        self.subj = None
        self.subj_has_data = False
        self.buffer = []

    def is_record_empty(self) -> bool:
        return not self.subj_has_data

    def write_data_property(self, predicate_id: str, value: Any, dtype: Optional[str]):
        if self.subj is None:
            return
        if dtype == DREPR_URI:
            value = URIRef(value)
        else:
            # to handle a bug in RDFlib that does not serialize integer properly.
            if (
                dtype == "http://www.w3.org/2001/XMLSchema#integer"
                or dtype == "http://www.w3.org/2001/XMLSchema#long"
                or dtype == "http://www.w3.org/2001/XMLSchema#int"
            ):
                value = int(float(value))
            value = Literal(value, datatype=dtype)

        self.subj_has_data = True
        if self.is_buffered:
            self.buffer.append((URIRef(predicate_id), value))
        else:
            assert self.subj is not None
            self.g.add((self.subj, URIRef(predicate_id), value))

    def write_object_property(
        self,
        predicate_id: str,
        object: SubjVal,
        is_subject_blank: bool,
        is_object_blank: bool,
        is_new_subj: bool,
    ):
        if self.subj is None:
            return
        object = self.written_records[object]
        self.subj_has_data = True
        if self.is_buffered:
            self.buffer.append((URIRef(predicate_id), object))
        else:
            assert self.subj is not None
            self.g.add((self.subj, URIRef(predicate_id), object))

    def write_to_string(self):
        return self.g.serialize(format="ttl")

    def write_to_file(self, filepath):
        self.g.serialize(filepath, format="ttl")



from __future__ import annotations

import importlib.util
from dataclasses import dataclass
from pathlib import Path
from threading import local
from typing import Annotated, Optional
from uuid import uuid4

import typer

from drepr.models.prelude import DRepr, OutputFormat
from drepr.planning.class_map_plan import ClassesMapExecutionPlan
from drepr.program_generation.main import FileOutput, MemoryOutput, Output, gen_program

app = typer.Typer(pretty_exceptions_short=True, pretty_exceptions_enable=False)


@dataclass
class ResourceInput:
    id: str
    file: Path

    @staticmethod
    def from_string(value: str) -> ResourceInput:
        lst = value.split("=", 1)
        if len(lst) != 2:
            raise Exception(f"Invalid resource format. Expect <id>`=`<file_path>")

        id, file = lst
        file = Path(file)
        if not file.exists():
            raise Exception(f"Resource file `{file}` does not exist")

        return ResourceInput(id, file)


@app.command()
def main(
    repr: Annotated[
        Path,
        typer.Option(
            help="A path to a file containing representation (support 2 formats: JSON & YML)",
            exists=True,
            dir_okay=False,
        ),
    ],
    resource: Annotated[
        list[str],
        typer.Option(
            help="file paths of resources in this format: <resource_id>=<file_path>",
        ),
    ],
    progfile: Annotated[
        Optional[Path],
        typer.Option(
            help="A path to a file to save the generated program", exists=False
        ),
    ] = None,
    outfile: Annotated[
        Optional[Path],
        typer.Option(
            help="A path to a file to save the transformed data", exists=False
        ),
    ] = None,
    format: Annotated[
        OutputFormat,
        typer.Option(
            help="The output format",
        ),
    ] = OutputFormat.TTL,
    tmpdir: Annotated[
        Path,
        typer.Option(
            help="A directory to save temporary files",
            default=Path("/tmp/drepr"),
        ),
    ] = Path("/tmp/drepr"),
    debuginfo: Annotated[
        bool, typer.Option(help="Whether to add debug information to the program")
    ] = False,
):
    parsed_resources = {
        (x := ResourceInput.from_string(r)).id: x.file for r in resource
    }
    parsed_repr = DRepr.parse_from_file(repr)
    exec_plan = ClassesMapExecutionPlan.create(parsed_repr)

    if outfile is not None:
        output = FileOutput(outfile, format)
    else:
        output = MemoryOutput(format)

    prog = gen_program(exec_plan.desc, exec_plan, output, debuginfo).to_python()
    cleanup = progfile is None
    if progfile is not None:
        with open(progfile, "w") as f:
            f.write(prog)
    else:
        tmpdir.mkdir(parents=True, exist_ok=True)
        unique_id = str(uuid4()).replace("-", "_")
        progfile = tmpdir / f"main_{unique_id}.py"
        progfile.write_text(prog)

    spec = importlib.util.spec_from_file_location("drepr_prog", progfile)
    assert spec is not None and spec.loader is not None
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    if outfile is not None:
        module.main(
            *[
                parsed_resources[r.id]
                for r in exec_plan.desc.resources
                if not r.is_preprocessing_output()
            ],
            outfile,
        )
    else:
        print(
            module.main(
                *[
                    parsed_resources[r.id]
                    for r in exec_plan.desc.resources
                    if not r.is_preprocessing_output()
                ]
            )
        )

    if cleanup:
        progfile.unlink()


if __name__ == "__main__":
    app()



from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass
from typing import Any, Optional, TypeAlias, cast

from drepr.models.prelude import (
    DREPR_URI,
    MISSING_VALUE_TYPE,
    Alignment,
    Attr,
    AttrId,
    Cardinality,
    ClassNode,
    DataNode,
    DRepr,
    EdgeId,
    LiteralNode,
    NodeId,
    Resource,
)
from drepr.models.sm import DREPR_BLANK
from drepr.planning.drepr_model_alignments import DReprModelAlignments
from drepr.planning.topological_sorting import topological_sorting


@dataclass
class ClassesMapExecutionPlan:
    desc: DRepr
    # read_plans: list[ReadPlan]
    # write_plan: WritePlan
    class_map_plans: list[ClassMapPlan]

    @classmethod
    def create(cls, desc: DRepr):
        desc = cls.rewrite_desc_for_preprocessing_output(desc)

        reversed_topo_orders = topological_sorting(desc.sm)
        alignments = DReprModelAlignments.create(desc)
        edges_optional = {e.edge_id: not e.is_required for e in desc.sm.edges.values()}
        edges_has_missing_values: dict[EdgeId, bool] = {}
        class_map_plans = []
        class2plan = {}

        # find subject attribute of each class
        class2subj = {}
        for class_id in reversed_topo_orders.topo_order:
            class2subj[class_id] = ClassMapPlan.find_subject(
                desc, class_id, class2subj, alignments
            )

        # figure out if an edge may contain missing values
        n_miss_edges = -1
        while n_miss_edges != len(edges_has_missing_values):
            n_miss_edges = len(edges_has_missing_values)
            for e in desc.sm.edges.values():
                if e.edge_id in edges_has_missing_values:
                    continue
                v = desc.sm.nodes[e.target_id]
                if isinstance(v, DataNode):
                    v_attr = desc.get_attr_by_id(v.attr_id)
                    edges_has_missing_values[e.edge_id] = (
                        len(v_attr.missing_values) > 0
                    ) or v_attr.path.has_optional_steps()
                elif isinstance(v, ClassNode):
                    if any(
                        ve.edge_id not in edges_has_missing_values
                        for ve in desc.sm.iter_outgoing_edges(v.node_id)
                        if not reversed_topo_orders.removed_outgoing_edges[ve.edge_id]
                    ):
                        continue

                    if v.is_blank_node(desc.sm):
                        # it can have missing values when:
                        # (1) at least one edge is mandatory & the edge have missing values
                        # (2) all edges are optional and all of them have missing values
                        have_missing_values = False

                        # (1) at least one edge is mandatory & the edge have missing values
                        for ve in desc.sm.iter_outgoing_edges(v.node_id):
                            if reversed_topo_orders.removed_outgoing_edges[ve.edge_id]:
                                continue

                            if (
                                not edges_optional[ve.edge_id]
                                and edges_has_missing_values[ve.edge_id]
                            ):
                                have_missing_values = True
                                break

                        # (2) all edges are optional and all of them have missing values
                        if not have_missing_values:
                            if all(
                                edges_optional[ve.edge_id]
                                and edges_has_missing_values[ve.edge_id]
                                for ve in desc.sm.iter_outgoing_edges(v.node_id)
                                if not reversed_topo_orders.removed_outgoing_edges[
                                    ve.edge_id
                                ]
                            ):
                                have_missing_values = True
                    else:
                        # it can have missing values when:
                        # (1) at least one edge is mandatory & the edge have missing values
                        # (2) the URI edge have missing values
                        have_missing_values = False

                        # (1) at least one edge is mandatory & the edge have missing values
                        for ve in desc.sm.iter_outgoing_edges(v.node_id):
                            if reversed_topo_orders.removed_outgoing_edges[ve.edge_id]:
                                continue

                            if (
                                not edges_optional[ve.edge_id]
                                and edges_has_missing_values[ve.edge_id]
                            ):
                                have_missing_values = True
                                break

                        # (2) the URI edge have missing values
                        if not have_missing_values:
                            ve_attr = desc.get_attr_by_id(class2subj[v.node_id])
                            if (
                                len(ve_attr.missing_values) > 0
                                or ve_attr.path.has_optional_steps()
                            ):
                                have_missing_values = True

                    edges_has_missing_values[e.edge_id] = have_missing_values
                else:
                    assert isinstance(v, LiteralNode)
                    edges_has_missing_values[e.edge_id] = False

        assert len(edges_has_missing_values) == len(desc.sm.edges) - sum(
            reversed_topo_orders.removed_outgoing_edges.values()
        ), f"{len(edges_has_missing_values)} == {len(desc.sm.edges)} - {sum(reversed_topo_orders.removed_outgoing_edges.values())}"

        # generate plans
        for class_id in reversed_topo_orders.topo_order:
            classplan = ClassMapPlan.create(
                desc,
                class_id,
                class2plan,
                class2subj,
                alignments,
                edges_optional,
                edges_has_missing_values,
                reversed_topo_orders.removed_outgoing_edges,
            )
            class2plan[class_id] = classplan
            class_map_plans.append(classplan)

        return ClassesMapExecutionPlan(desc, class_map_plans)

    @classmethod
    def rewrite_desc_for_preprocessing_output(cls, desc: DRepr):
        if all(not pre.is_output_new_data() for pre in desc.preprocessing):
            return desc

        desc = deepcopy(desc)
        assert all(not r.is_preprocessing_output() for r in desc.resources)
        for pre in desc.preprocessing:
            if not pre.is_output_new_data():
                continue

            attr = pre.get_new_data_attribute("")
            newresource = Resource.create_preprocessing_output(attr.id)
            attr.resource_id = newresource.id

            assert not desc.has_attr(
                attr.id
            ), f"Preprocessing attempts to write to an existing attribute: {attr.id}. Please change the output attribute id"
            desc.attrs.append(attr)
            desc.add_resource(newresource)

        return desc


@dataclass
class ClassMapPlan:
    class_id: str
    subject: Subject
    data_props: list[DataProp]
    literal_props: list[LiteralProp]
    object_props: list[ObjectProp]
    buffered_object_props: list[ObjectProp]

    @staticmethod
    def create(
        desc: DRepr,
        class_id: NodeId,
        class2plan: dict[NodeId, ClassMapPlan],
        class2subj: dict[NodeId, AttrId],
        inference: DReprModelAlignments,
        edges_optional: dict[EdgeId, bool],
        edges_missing_values: dict[EdgeId, bool],
        removed_edges: dict[EdgeId, bool],
    ):
        subj = class2subj[class_id]
        uri_dnode: Optional[DataNode] = None
        blank_dnode: Optional[DataNode] = None
        for e in desc.sm.iter_outgoing_edges(class_id):
            e_abs_uri = e.get_abs_iri(desc.sm)
            if e_abs_uri == DREPR_URI:
                assert uri_dnode is None
                tmp = desc.sm.nodes[e.target_id]
                assert isinstance(tmp, DataNode)
                uri_dnode = tmp
            elif e_abs_uri == DREPR_BLANK:
                assert blank_dnode is None
                tmp = desc.sm.nodes[e.target_id]
                assert isinstance(tmp, DataNode)
                blank_dnode = tmp

        assert not (uri_dnode is not None and blank_dnode is not None)

        # generate other properties
        literal_props = []
        data_props: list[DataProp] = []
        object_props = []
        buffered_object_props = []

        for e in desc.sm.iter_outgoing_edges(class_id):
            target = desc.sm.nodes[e.target_id]
            if isinstance(target, DataNode):
                attribute = desc.get_attr_by_id(target.attr_id)
                e_abs_iri = e.get_abs_iri(desc.sm)
                if e_abs_iri != DREPR_URI and e_abs_iri != DREPR_BLANK:
                    alignments = inference.get_alignments(subj, target.attr_id)
                    alignments_cardinality = inference.estimate_cardinality(alignments)

                    if attribute.value_type.is_list():
                        if alignments_cardinality.is_one_to_star():
                            alignments_cardinality = Cardinality.OneToMany
                        else:
                            assert (
                                alignments_cardinality.is_many_to_star()
                            ), alignments_cardinality
                            alignments_cardinality = Cardinality.ManyToMany

                    data_props.append(
                        DataProp(
                            alignments=alignments,
                            alignments_cardinality=inference.estimate_cardinality(
                                alignments
                            ),
                            predicate=e.get_abs_iri(desc.sm),
                            attr=attribute,
                            is_optional=edges_optional[e.edge_id],
                            missing_values=set(attribute.missing_values),
                            missing_path=attribute.path.has_optional_steps(),
                            datatype=(target.data_type),
                        )
                    )
            elif isinstance(target, LiteralNode):
                literal_props.append(
                    LiteralProp(
                        predicate=e.get_abs_iri(desc.sm),
                        value=target.value,
                        datatype=target.data_type,
                    )
                )
            elif isinstance(target, ClassNode):
                attribute = desc.get_attr_by_id(class2subj[e.target_id])

                alignments = inference.get_alignments(subj, attribute.id)

                if target.is_blank_node(desc.sm):
                    if any(
                        etmp.get_abs_iri(desc.sm) == DREPR_BLANK
                        for etmp in desc.sm.iter_outgoing_edges(target.node_id)
                    ):
                        # ensure that it is consistent
                        assert (
                            target.node_id in class2plan
                            and isinstance(
                                (
                                    targetclsplansubj := class2plan[
                                        target.node_id
                                    ].subject
                                ),
                                BlankSubject,
                            )
                            and targetclsplansubj.use_attr_value
                        )
                        use_attr_value = True
                    else:
                        use_attr_value = False
                    prop = BlankObject(
                        attr=attribute,
                        alignments_cardinality=inference.estimate_cardinality(
                            alignments
                        ),
                        alignments=alignments,
                        predicate=e.get_abs_iri(desc.sm),
                        class_id=class_id,
                        is_optional=edges_optional[e.edge_id],
                        can_target_missing=edges_missing_values[e.edge_id],
                        use_attr_value=use_attr_value,
                    )
                else:
                    prop = IDObject(
                        attr=attribute,
                        alignments_cardinality=inference.estimate_cardinality(
                            alignments
                        ),
                        alignments=alignments,
                        predicate=e.get_abs_iri(desc.sm),
                        class_id=class_id,
                        is_optional=edges_optional[e.edge_id],
                        can_target_missing=edges_missing_values[e.edge_id],
                        missing_values=set(attribute.missing_values),
                    )

                if removed_edges[e.edge_id]:
                    buffered_object_props.append(prop)
                else:
                    object_props.append(prop)

        subj_attr = desc.get_attr_by_id(subj)

        if uri_dnode is None:
            if blank_dnode is not None:
                assert blank_dnode.attr_id == subj
            subject = BlankSubject(
                attr=subj_attr,
                use_attr_value=blank_dnode is not None,
                missing_values=set(subj_attr.missing_values),
            )
        else:
            # get missing values from the real subjects
            missing_values = set(subj_attr.missing_values)
            (uri_dnode_inedge,) = [
                e
                for e in desc.sm.get_edges_between_nodes(class_id, uri_dnode.node_id)
                if e.get_abs_iri(desc.sm) == DREPR_URI
            ]

            if uri_dnode.attr_id == subj:
                subject = InternalIDSubject(
                    attr=subj_attr,
                    is_optional=edges_optional[uri_dnode_inedge.edge_id],
                    missing_values=missing_values,
                )
            else:
                subject = ExternalIDSubject(
                    attr=subj_attr,
                    is_optional=edges_optional[uri_dnode_inedge.edge_id],
                    missing_values=missing_values,
                )

        return ClassMapPlan(
            class_id=class_id,
            subject=subject,
            data_props=data_props,
            literal_props=literal_props,
            object_props=object_props,
            buffered_object_props=buffered_object_props,
        )

    @staticmethod
    def find_subject(
        desc: DRepr,
        class_id: NodeId,
        class2subj: dict[NodeId, AttrId],
        desc_aligns: DReprModelAlignments,
    ):
        """
        Find the subject of the class among the attributes of the class.

        The subject has *-to-one relationship with other attributes.
        """
        # get data nodes, attributes, and the attribute that contains URIs of the class
        data_nodes: list[DataNode] = []
        attrs: list[AttrId] = []
        uri_attr: Optional[AttrId] = None
        blank_attr: Optional[AttrId] = None

        for e in desc.sm.iter_outgoing_edges(class_id):
            target = desc.sm.nodes[e.target_id]

            if isinstance(target, DataNode):
                data_nodes.append(target)
                attrs.append(target.attr_id)

                e_abs_iri = e.get_abs_iri(desc.sm)
                if e_abs_iri == DREPR_URI:
                    assert uri_attr is None
                    uri_attr = target.attr_id
                elif e_abs_iri == DREPR_BLANK:
                    assert blank_attr is None
                    blank_attr = target.attr_id

        assert not (uri_attr is not None and blank_attr is not None)

        if uri_attr is not None:
            # always use the URI attribute because there can be a property that is a list
            return uri_attr

        if blank_attr is not None:
            # always use the blank attribute because there can be a property that is a list
            return blank_attr

        # if the subject attribute is provided, then, we will use it.
        subjs = []
        for u in data_nodes:
            if any(
                e.is_subject
                for e in desc.sm.get_edges_between_nodes(class_id, u.node_id)
            ):
                subjs.append(u.attr_id)

        if len(subjs) == 0:
            if len(attrs) == 0:
                # there is a special case where the class has no data node, but only links to other classes
                # we need to get the subject from the other classes
                other_attrs = []
                for e in desc.sm.iter_outgoing_edges(class_id):
                    target = desc.sm.nodes[e.target_id]
                    if isinstance(target, ClassNode):
                        # we must have inferred the subject of the target class before (because of the topological sorting)
                        assert target.node_id in class2subj
                        target_subj = class2subj[target.node_id]
                        other_attrs.append(target_subj)
                subjs = desc_aligns.infer_subject(other_attrs)
            else:
                # invoke the inference to find the subject attribute
                subjs = desc_aligns.infer_subject(attrs)

        if len(subjs) == 0:
            raise Exception(
                "There is no subject attribute of class: {} ({}). Users need to specify it explicitly".format(
                    cast(ClassNode, desc.sm.nodes[class_id]).label, class_id
                )
            )

        return ClassMapPlan.select_subject(desc, class_id, subjs, attrs)

    @staticmethod
    def select_subject(
        desc: DRepr,
        class_id: NodeId,
        subjs: list[AttrId],
        attrs: list[AttrId],
    ) -> AttrId:
        return subjs[0]


@dataclass
class BlankSubject:
    attr: Attr
    # whether to use the attribute value as the value of the subject
    use_attr_value: bool
    # only works when use_attr_value is True
    missing_values: set[MISSING_VALUE_TYPE]


@dataclass
class InternalIDSubject:
    attr: Attr
    is_optional: bool
    missing_values: set[MISSING_VALUE_TYPE]


@dataclass
class ExternalIDSubject:
    attr: Attr
    is_optional: bool
    missing_values: set[MISSING_VALUE_TYPE]


Subject: TypeAlias = BlankSubject | InternalIDSubject | ExternalIDSubject


@dataclass
class DataProp:
    alignments: list[Alignment]
    alignments_cardinality: Cardinality
    predicate: str
    attr: Attr
    is_optional: bool
    missing_values: set[MISSING_VALUE_TYPE]
    missing_path: bool
    datatype: Optional[str]

    @property
    def can_target_missing(self):
        return len(self.missing_values) > 0 or self.missing_path


@dataclass
class LiteralProp:
    predicate: str
    value: Any
    datatype: Optional[str]


@dataclass
class BlankObject:
    attr: Attr
    alignments: list[Alignment]
    alignments_cardinality: Cardinality
    predicate: str
    class_id: NodeId
    is_optional: bool
    # whether an instance of the target class can be missing
    can_target_missing: bool
    # whether to use the attribute value as the value of the blank object
    use_attr_value: bool


@dataclass
class IDObject:
    attr: Attr
    alignments: list[Alignment]
    alignments_cardinality: Cardinality
    predicate: str
    class_id: NodeId
    is_optional: bool
    # whether an instance of the target class can be missing
    can_target_missing: bool
    missing_values: set[MISSING_VALUE_TYPE]


ObjectProp = BlankObject | IDObject



def safe_item_getter(collection, item, msg):
    try:
        return collection[item]
    except Exception as e:
        raise Exception(msg) from e


def safe_range(collection, item, msg):
    try:
        return range(collection, item, item)
    except Exception as e:
        raise Exception(msg) from e


def safe_len(collection, msg):
    try:
        return len(collection)
    except Exception as e:
        raise Exception(msg) from e



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

import drepr.models.path as dpath
from codegen.models import AST, Program, Var, expr
from codegen.models.var import DeferredVar
from drepr.models.prelude import (
    Attr,
    Context,
    DRepr,
    PMap,
    PreprocessingType,
    Resource,
    Sorted,
    ValueType,
)
from drepr.program_generation.alignment_fn import PathAccessor
from drepr.program_generation.predefined_fn import DReprPredefinedFn
from drepr.program_generation.program_space import VarSpace
from drepr.utils.udf import SourceTree, UDFParsedResult, UDFParser

PreprocessingId = int


@dataclass
class NormedUserDefinedFn:
    name: str
    fnvar: expr.Expr
    udf: UDFParsedResult
    use_context: bool


class GenPreprocessing:
    """Generate preprocessing code for the given D-REPR."""

    def __init__(self, program: Program, desc: DRepr, call_preproc_ast: AST):
        self.program = program
        self.call_preproc_ast = call_preproc_ast
        self.desc = desc

        self.user_defined_fn: dict[PreprocessingId, NormedUserDefinedFn] = {}

    def generate(self):
        for i in range(len(self.desc.preprocessing)):
            self._generate_preprocessing(i)

    def _generate_preprocessing(self, prepro_id: PreprocessingId):
        preprocessing = self.desc.preprocessing[prepro_id]
        genfn_name = f"preprocess_{prepro_id}"

        self.program.root.linebreak()
        prepro_fn = self.program.root.func(
            genfn_name,
            [
                DeferredVar(
                    name="resource_data",
                    key=VarSpace.resource_data(preprocessing.value.resource_id),
                ),
            ],
        )
        self.program.root.linebreak()

        if preprocessing.type == PreprocessingType.pmap:
            value = preprocessing.value
            assert isinstance(value, PMap)

            # generate the necessary function (user defined fn & our preprocessing fn)
            self._init_user_defined_fn(prepro_id, value.code)
            self._generate_preprocessing_pmap(prepro_id, prepro_fn, value)

            prepro_invoke_expr = expr.ExprFuncCall(
                expr.ExprIdent(genfn_name),
                [
                    expr.ExprVar(
                        self.program.get_var(
                            key=VarSpace.resource_data(preprocessing.value.resource_id),
                            at=self.call_preproc_ast.next_child_id(),
                        )
                    ),
                ],
            )
            # call preprocessing fn in the main program and assign if needed
            if value.output is not None:
                prepro_resource_id = Resource.get_preprocessing_output_id(value.output)
                self.call_preproc_ast.assign(
                    DeferredVar(
                        name=f"resource_data_{prepro_resource_id}",
                        key=VarSpace.resource_data(prepro_resource_id),
                    ),
                    prepro_invoke_expr,
                )
            else:
                self.call_preproc_ast.expr(prepro_invoke_expr)
            return None

        raise NotImplementedError(preprocessing.type)

    def _generate_preprocessing_pmap(self, prepro_id: int, prepro_fn: AST, value: PMap):
        if not value.change_structure:
            if value.output is not None:
                prepro_resource_id = Resource.get_preprocessing_output_id(value.output)
                # create a variable to store the preprocess results
                output_attr_id = value.output
                output_attr = DeferredVar(
                    name=output_attr_id, key=VarSpace.preprocessing_output(prepro_id)
                )

                # we use a special class called AttributeData that can store data as if it is in the resource data (same index, etc -- but the
                # real location in memory is different)
                self.program.import_("drepr.utils.attr_data.AttributeData", True)
                prepro_fn.assign(
                    output_attr,
                    expr.ExprFuncCall(
                        expr.ExprIdent("AttributeData.from_raw_path"),
                        [
                            expr.ExprConstant(value.path.to_lang_format()),
                        ],
                    ),
                )

                def on_step(
                    ast: AST,
                    dim: int,
                    collection: expr.Expr,
                    key: expr.Expr,
                    result: DeferredVar | Var,
                ):
                    # everytime we step into a new dimension, we step into the output variable as well.
                    if dim == 0:
                        # this is better than `output_attr.get_var()` as we can ensure the scope is correct
                        output_collection = expr.ExprVar(
                            self.program.get_var(
                                key=output_attr.key, at=ast.next_child_id()
                            )
                        )
                    else:
                        output_collection = expr.ExprVar(
                            self.program.get_var(
                                key=VarSpace.attr_value_dim(
                                    prepro_resource_id, output_attr_id, dim - 1
                                ),
                                at=ast.next_child_id(),
                            )
                        )

                    itemvalue = DeferredVar(
                        name=f"{output_attr_id}_value_{dim}",
                        key=VarSpace.attr_value_dim(
                            prepro_resource_id, output_attr_id, dim
                        ),
                    )
                    ast.assign(
                        itemvalue, DReprPredefinedFn.item_getter(output_collection, key)
                    )

                on_step_callback = on_step
            else:
                on_step_callback = None

            # the idea is to loop through the path without the last index, get the value, apply the function, and set
            # the value of the last index to the new value
            pseudo_attr = Attr(
                id=f"preproc_{prepro_id}_path",
                resource_id=value.resource_id,
                path=value.path,
                missing_values=[],
                unique=False,
                sorted=Sorted.Null,
                value_type=ValueType.UnspecifiedSingle,
            )
            ast = PathAccessor(self.program).iterate_elements(
                ast=prepro_fn, attr=pseudo_attr, on_step_callback=on_step_callback
            )

            # get item value & item context
            item_value = self.program.get_var(
                key=VarSpace.attr_value_dim(
                    pseudo_attr.resource_id,
                    pseudo_attr.id,
                    len(pseudo_attr.path.steps) - 1,
                ),
                at=ast.next_child_id(),
            )
            if self.user_defined_fn[prepro_id].use_context:
                context_index: list[expr.Expr] = []
                for dim in range(len(pseudo_attr.path.steps)):
                    step = pseudo_attr.path.steps[dim]
                    if isinstance(step, dpath.IndexExpr):
                        if isinstance(step.val, dpath.Expr):
                            # I don't know about recursive path expression yet -- what usecase and how to use them -- so I can't implement.
                            raise Exception(
                                f"Recursive path expression is not supported yet. Please raise a ticket to notify us for future support! Found: {step.vals}"
                            )
                        context_index.append(expr.ExprConstant(step.val))
                    else:
                        context_index.append(
                            expr.ExprVar(
                                self.program.get_var(
                                    key=VarSpace.attr_index_dim(
                                        pseudo_attr.resource_id, pseudo_attr.id, dim
                                    ),
                                    at=ast.next_child_id(),
                                )
                            )
                        )

                self.program.import_(
                    "drepr.program_generation.preprocessing.ContextImpl", True
                )
                item_context = expr.ExprFuncCall(
                    expr.ExprIdent("ContextImpl"),
                    [
                        expr.ExprVar(
                            self.program.get_var(
                                key=VarSpace.resource_data(value.resource_id),
                                at=prepro_fn.next_child_id(),
                            )
                        ),
                        DReprPredefinedFn.tuple(context_index),
                    ],
                )
            else:
                item_context = None

            # then we call the user defined fn to get the new item value
            new_item_value = self._call_user_defined_fn(
                prepro_id,
                ast,
                expr.ExprVar(item_value),
                item_context,
            )

            # get the parent item value & index to assign the new value
            # if output new data, we need to retrieve the parent item value from correct location
            if value.output is not None:
                if len(pseudo_attr.path.steps) > 1:
                    parent_item_value = self.program.get_var(
                        key=VarSpace.attr_value_dim(
                            prepro_resource_id,
                            output_attr_id,
                            len(pseudo_attr.path.steps) - 2,
                        ),
                        at=ast.next_child_id(),
                    )
                else:
                    # this is better than `output_attr.get_var()` as we can ensure the scope is correct
                    parent_item_value = self.program.get_var(
                        key=output_attr.key,
                        at=ast.next_child_id(),
                    )
            else:
                if len(pseudo_attr.path.steps) > 1:
                    parent_item_value = self.program.get_var(
                        key=VarSpace.attr_value_dim(
                            pseudo_attr.resource_id,
                            pseudo_attr.id,
                            len(pseudo_attr.path.steps) - 2,
                        ),
                        at=ast.next_child_id(),
                    )
                else:
                    parent_item_value = self.program.get_var(
                        key=VarSpace.resource_data(pseudo_attr.resource_id),
                        at=ast.next_child_id(),
                    )

            if isinstance(pseudo_attr.path.steps[-1], dpath.IndexExpr):
                parent_item_index = expr.ExprConstant(pseudo_attr.path.steps[-1].val)
            else:
                parent_item_index = expr.ExprVar(
                    self.program.get_var(
                        key=VarSpace.attr_index_dim(
                            pseudo_attr.resource_id,
                            pseudo_attr.id,
                            len(pseudo_attr.path.steps) - 1,
                        ),
                        at=ast.next_child_id(),
                    )
                )

            # then, we set the new item value to the parent item value
            ast.expr(
                DReprPredefinedFn.item_setter(
                    expr.ExprVar(parent_item_value),
                    parent_item_index,
                    new_item_value,
                )
            )

            if value.output is not None:
                prepro_fn.return_(
                    expr.ExprVar(
                        self.program.get_var(
                            key=output_attr.key,
                            at=prepro_fn.next_child_id(),
                        )
                    )
                )
        else:
            # haven't considered the case of changing structure yet
            raise NotImplementedError()

    def _init_user_defined_fn(self, prepro_id: PreprocessingId, code: str):
        """First, import statements are moved to the top of the file. The rest of the code can be either
        wrapped in a function (as expected in DRepr design), or directly embedded whenever it is used.

        The later option (embedding code) yields more performance, but it is harder because of potential variable name conflicts.
        To implement the embedding code, we need to parse the code to find all variables that are used in the code and
        ensure that they are not used or overwrite previous variables (potential renaming may require). Also, we need to rewrite
        the return statement.
        """
        # detect indentation & remove it
        parsed_udf = UDFParser(code).parse(["context"])

        # now create a function containing the user-defined function
        fnname = f"preproc_{prepro_id}_customfn"
        fnargs = [
            DeferredVar(
                name="value",
                key=VarSpace.preprocessing_udf_value(
                    self.desc.preprocessing[prepro_id].value.resource_id
                ),
                force_name="value",
            )
        ]
        use_context = "context" in parsed_udf.monitor_variables

        if use_context:
            fnargs.append(
                DeferredVar(
                    name="context",
                    key=VarSpace.preprocessing_udf_context(
                        self.desc.preprocessing[prepro_id].value.resource_id
                    ),
                    force_name="context",
                )
            )

        if len(parsed_udf.imports) > 0:
            # create a function that will be used to create the user-defined function
            create_udf = self.program.root.func("get_" + fnname, [])
            self.program.root.linebreak()

            for import_stmt in parsed_udf.imports:
                create_udf.python_stmt(import_stmt)

            inner_udf = create_udf.func(fnname, fnargs)
            create_udf.return_(expr.ExprIdent(fnname))

            def insert_source_tree(ast: AST, tree: SourceTree):
                assert tree.node != ""
                ast = ast.python_stmt(tree.node)
                for child in tree.children:
                    insert_source_tree(ast, child)

            assert parsed_udf.source_tree.node == ""
            assert len(parsed_udf.source_tree.children) > 0
            for child in parsed_udf.source_tree.children:
                insert_source_tree(inner_udf, child)

            # now we create the user-defined function
            fnvar = DeferredVar(fnname, key=VarSpace.preprocessing_udf(fnname))
            self.program.root.assign(
                fnvar,
                expr.ExprFuncCall(expr.ExprIdent("get_" + fnname), []),
            )
            fnvar = expr.ExprVar(fnvar.get_var())
        else:
            inner_udf = self.program.root.func(fnname, fnargs)
            self.program.root.linebreak()

            def insert_source_tree(ast: AST, tree: SourceTree):
                assert tree.node != ""
                ast = ast.python_stmt(tree.node)
                for child in tree.children:
                    insert_source_tree(ast, child)

            assert parsed_udf.source_tree.node == ""
            assert len(parsed_udf.source_tree.children) > 0
            for child in parsed_udf.source_tree.children:
                insert_source_tree(inner_udf, child)

            # now we create the user-defined function
            fnvar = expr.ExprIdent(fnname)

        self.user_defined_fn[prepro_id] = NormedUserDefinedFn(
            name=fnname, udf=parsed_udf, use_context=use_context, fnvar=fnvar
        )

    def _call_user_defined_fn(
        self,
        prepro_id: PreprocessingId,
        ast: AST,
        value: expr.Expr,
        context: Optional[expr.Expr],
    ) -> expr.Expr:
        """Call the user-defined function"""
        return expr.ExprFuncCall(
            self.user_defined_fn[prepro_id].fnvar,
            [value, context] if context is not None else [value],
        )


class ContextImpl(Context):
    def __init__(self, resource_data, index: tuple):
        self.resource_data = resource_data
        self.index = index

    def get_index(self) -> tuple:
        return self.index

    def get_value(self, index: tuple):
        ptr = self.resource_data
        for i in index:
            ptr = ptr[i]
        return ptr


class GenerateDataStorage:
    pass



from drepr.utils.misc import get_abs_iri


class NamespaceMixin:
    prefixes: dict[str, str]

    @classmethod
    def is_rel_iri(cls, iri: str) -> bool:
        return iri.find("://") == -1 and iri.find(":") != -1

    def get_rel_iri(self, abs_iri: str) -> str:
        """Convert an absolute IRI to a relative IRI."""
        assert not self.is_rel_iri(abs_iri)
        for prefix, uri in self.prefixes.items():
            if abs_iri.startswith(uri):
                return f"{prefix}:{abs_iri.replace(uri, '')}"
        raise ValueError(
            "Cannot create relative IRI because there is no suitable prefix"
        )

    def get_abs_iri(self, rel_iri: str) -> str:
        """Convert a relative IRI to an absolute IRI."""
        assert self.is_rel_iri(rel_iri)
        return get_abs_iri(self.prefixes, rel_iri)



from __future__ import annotations

import ast
import re
from dataclasses import dataclass
from typing import Optional


@dataclass
class SourceTree:
    node: str
    children: list[SourceTree]

    def get_simplified_dict(self):
        children = [
            child.get_simplified_dict() if len(child.children) > 0 else child.node
            for child in self.children
        ]

        if self.node == "":
            return children

        return {
            "node": self.node,
            "children": children,
        }


@dataclass
class UDFParsedResult:
    # import statements
    imports: list[str]
    # set of variables that we opt to monitor during parsing (e.g., we want to detect a variable named context)
    monitor_variables: set[str]
    # the source code of the UDF -- without import statements
    source_tree: SourceTree


class UDFParser:

    def __init__(self, source_code: str):
        if source_code.strip() == "":
            raise ValueError(f"Cannot parse an empty code")

        self.source_code = self.remove_prefix_spaces(source_code)
        self.source_code_lines = self.source_code.split("\n")
        self.source_tree = ast.parse(self.source_code)

    def parse(self, monitor_vars: Optional[list[str]] = None) -> UDFParsedResult:
        imports = []
        tree = SourceTree("", [])
        for stmt in self.source_tree.body:
            tree.children.extend(self._parse_ast(stmt, imports))

        found_vars = set()
        if monitor_vars is not None and len(monitor_vars) > 0:
            # we need to find all variables that we want to monitor
            for node in ast.walk(self.source_tree):
                if isinstance(node, ast.Name) and node.id in monitor_vars:
                    found_vars.add(node.id)

        return UDFParsedResult(
            imports=imports,
            monitor_variables=found_vars,
            source_tree=tree,
        )

    def _parse_ast(self, tree: ast.AST, imports: list[str]) -> list[SourceTree]:
        if isinstance(tree, (ast.Import, ast.ImportFrom)):
            imports.append(self._get_node_code(tree))
            return []

        if isinstance(
            tree,
            (ast.Expr, ast.Return, ast.Yield, ast.YieldFrom, ast.Assign, ast.Assert),
        ):
            return [SourceTree(self._get_node_code(tree), [])]

        if isinstance(tree, ast.If):
            content = f"if {self._get_node_code(tree.test)}"
            out = [SourceTree(content, [])]
            for stmt in tree.body:
                out[0].children.extend(self._parse_ast(stmt, imports))
            if len(tree.orelse) > 0:
                out.append(SourceTree("else:", []))
                for stmt in tree.orelse:
                    out[1].children.extend(self._parse_ast(stmt, imports))
            return out

        if isinstance(tree, ast.For):
            content = f"for {self._get_node_code(tree.target)}{self._get_node_code(tree.iter)}"
            out = [SourceTree(content, [])]
            for stmt in tree.body:
                out[0].children.extend(self._parse_ast(stmt, imports))
            if len(tree.orelse) > 0:
                out.append(SourceTree("else:", []))
                for stmt in tree.orelse:
                    out[1].children.extend(self._parse_ast(stmt, imports))
            return out
        raise NotImplementedError()

    def _get_node_code(self, node: ast.AST) -> str:
        lines = self.source_code_lines[node.lineno - 1 : node.end_lineno]
        lines[0] = lines[0][node.col_offset :]
        lines[-1] = lines[-1][: node.end_col_offset]
        return "\n".join(lines)

    def remove_prefix_spaces(self, code: str) -> str:
        lines = [x.rstrip() for x in code.splitlines()]
        non_empty_line_no = next(i for i in range(len(lines)) if lines[i] != "")
        lines = lines[non_empty_line_no:]
        assert len(lines) > 0

        m = re.match(r"^([ \t]*)", lines[0])
        assert m is not None
        indentation = m.group(1)

        if not all(x.startswith(indentation) or x.strip() == "" for x in lines):
            raise ValueError(
                f"The code has inconsistent prefix spaces. The first line has {indentation} spaces, but the following lines do not have the same prefix spaces"
            )

        return "\n".join(x[len(indentation) :] for x in lines)



import re
from typing import List

from drepr.models.align import (
    AlignedStep,
    Alignment,
    AlignmentType,
    AutoAlignment,
    RangeAlignment,
    ValueAlignment,
)
from drepr.utils.validator import InputError, Validator


class AlignParser:
    """
    Alignments are defined as a list

    ```
    - type: <alignment_type>
      # other properties
    ```

    If <alignment_type> is `dimension`, we have two schemas:
        1. align by one dimension

        ```
        - type: dimension
          value: <source_attr_id>:<step_index> <-> <target_attr_id>:<step_index>
        ```

        2. align by multiple dimensions

        ```
        - type: dimension
          source: <attr_id>
          target: <attr_id>
          aligned_dims:
            - { source: <step_index>, target: <step_index> }
            - { source: <step_index>, target: <step_index> }
            ...
        ```

    If <alignment_type> is `value`, then the schema is
        ```
        - type: value
          source: <attr_id>
          target: <attr_id>
        ```

    If <alignment_type> is `auto`, then the schema is:
        ```
        - type: auto
          attributes: <list of attributes to align>
        ```

        The attributes are optional, if not provided, the system will try to align all attributes automatically based
        on the dimension
    """

    ALIGNMENT_TYPES_VALUES = {
        x.value if x != AlignmentType.Range else "dimension" for x in AlignmentType
    }
    REG_DALIGN = re.compile(r"^([^:]+)(?::(\d+))? *<-> *([^:]+)(?::(\d+))?$")
    DALIGN_KEYS = {"type", "source", "target", "aligned_dims"}
    DALIGN_ALIGNED_DIMS_KEYS = {"source", "target"}
    VALIGN_KEYS = {"type", "source", "target"}

    @classmethod
    def parse(cls, alignments: List[dict]) -> List[Alignment]:
        Validator.must_be_list(alignments, "Parsing alignments")
        aligns = []

        for i, align in enumerate(alignments):
            trace = f"Parsing an alignment at index {i}"
            Validator.must_be_dict(align, trace)
            Validator.must_have(align, "type", trace)

            align_type = AlignmentType(
                align["type"]
                if align["type"] != "dimension"
                else "range" if align["type"] != "auto" else "auto"
            )
            if align_type == AlignmentType.Range:
                aligns.append(cls.parse_range_align(align, trace))
            elif align_type == AlignmentType.Value:
                aligns.append(cls.parse_value_align(align, trace))
            elif align_type == AlignmentType.Auto:
                aligns.append(cls.parse_auto_align(align, trace))
            else:
                raise NotImplemented(
                    f"{trace}\nERROR: not implement parser for alignment type: {align_type}"
                )

        return aligns

    @classmethod
    def parse_range_align(cls, conf: dict, parse_trace: str) -> RangeAlignment:
        if "value" in conf:
            Validator.must_be_str(
                conf["value"],
                f"{parse_trace}\nParsing property `value` of the alignment",
            )
            m = cls.REG_DALIGN.match(conf["value"])
            if m is None:
                raise InputError(
                    f"{parse_trace}\nERROR: invalid `value` property, its format is invalid"
                )

            source = m.group(1).strip()
            target = m.group(3).strip()
            if m.group(2) is None or m.group(4) is None:
                if not (m.group(2) is None and m.group(4) is None):
                    raise InputError(
                        f"{parse_trace}\nERROR: invalid `value` property, its format is invalid"
                    )
                return RangeAlignment(source, target, [])

            return RangeAlignment(
                source, target, [AlignedStep(int(m.group(2)), int(m.group(4)))]
            )

        Validator.must_be_subset(
            cls.DALIGN_KEYS, conf.keys(), "properties of alignment", parse_trace
        )
        aligned_dims = []
        for i, adim in enumerate(conf["aligned_dims"]):
            Validator.must_be_subset(
                cls.DALIGN_ALIGNED_DIMS_KEYS,
                adim.keys(),
                "properties of aligned_dims",
                parse_trace,
            )
            for key in cls.DALIGN_ALIGNED_DIMS_KEYS:
                Validator.must_be_int(
                    adim[key],
                    f"{parse_trace}\nParsing property `{key}` of aligned_dim at position {i}",
                )
            aligned_dims.append(AlignedStep(adim["source"], adim["target"]))
        return RangeAlignment(conf["source"], conf["target"], aligned_dims)

    @classmethod
    def parse_value_align(cls, conf: dict, parse_trace: str) -> ValueAlignment:
        Validator.must_be_subset(
            cls.VALIGN_KEYS, conf.keys(), "properties of alignment", parse_trace
        )
        for key in cls.VALIGN_KEYS:
            Validator.must_be_str(
                conf[key], f"{parse_trace}\nParsing property `{key}` of value alignment"
            )

        return ValueAlignment(conf["source"], conf["target"])

    @classmethod
    def parse_auto_align(cls, conf: dict, parse_trace: str) -> AutoAlignment:
        Validator.must_be_subset(
            {"type", "attributes"}, conf.keys(), "properties of alignment", parse_trace
        )

        Validator.must_be_str(
            conf["type"],
            f"{parse_trace}\nParsing property `{'type'}` of value alignment",
        )
        if "attributes" in conf:
            Validator.must_be_list_str(
                conf["attributes"],
                f"{parse_trace}\nParsing property `{'attributes'}` of value alignment",
            )

        return AutoAlignment(conf.get("attributes", None))



from __future__ import annotations

import importlib.util
from pathlib import Path
from typing import Optional
from uuid import uuid4

from drepr.models.prelude import DRepr, OutputFormat
from drepr.planning.class_map_plan import ClassesMapExecutionPlan
from drepr.program_generation.main import FileOutput, MemoryOutput, gen_program


def convert(
    repr: DRepr | Path,
    resources: dict[str, Path],
    progfile: Optional[Path] = None,
    outfile: Optional[Path] = None,
    format: OutputFormat = OutputFormat.TTL,
    tmpdir: Path = Path("/tmp/drepr"),
    debuginfo: bool = False,
    cleanup: bool = True,
):
    if not isinstance(repr, DRepr):
        repr = DRepr.parse_from_file(repr)

    exec_plan = ClassesMapExecutionPlan.create(repr)

    if outfile is not None:
        output = FileOutput(outfile, format)
    else:
        output = MemoryOutput(format)

    prog = gen_program(exec_plan.desc, exec_plan, output, debuginfo).to_python()
    cleanup = progfile is None
    if progfile is not None:
        with open(progfile, "w") as f:
            f.write(prog)
    else:
        tmpdir.mkdir(parents=True, exist_ok=True)
        unique_id = str(uuid4()).replace("-", "_")
        progfile = tmpdir / f"main_{unique_id}.py"
        progfile.write_text(prog)

    spec = importlib.util.spec_from_file_location("drepr_prog", progfile)
    assert spec is not None and spec.loader is not None
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    if outfile is not None:
        output = module.main(
            *[
                resources[r.id]
                for r in exec_plan.desc.resources
                if not r.is_preprocessing_output()
            ],
            outfile,
        )
    else:
        output = module.main(
            *[
                resources[r.id]
                for r in exec_plan.desc.resources
                if not r.is_preprocessing_output()
            ]
        )

    if cleanup:
        progfile.unlink()

    return output
